#include <cstdio>
#include <vector>

using namespace std;

#include "o65linker.hh"
#include "msginsert.hh"
#include "dataarea.hh"
#include "space.hh"
#include "argh.hh"

void MessageLinkingModules(unsigned count)
{
    //printf("Linking %u modules\n", count);
}

void MessageLoadingItem(const string& header)
{
    //printf("Loading %s\n", header.c_str());
}

void MessageWorking()
{
}

void MessageDone()
{
    //printf("- done\n");
}

int main(int argc, const char* const *argv)
{
    ParamHandler Argh;

    Argh.AddLong("help",       'h').SetBool().SetDesc("This help");
    Argh.AddLong("version",    500).SetBool().SetDesc("Displays version information");
    Argh.AddString(            'o').SetDesc("Place the output into <file>", "<file>");
    
    Argh.StartParse(argc, argv);
    
    std::vector<std::string> files;

    std::FILE *output = NULL;
    std::string outfn;

    for(;;)
    {      
        switch(Argh.GetParam())
        {                      
            case -1: goto EndArgList;
            
            case 500: /* version */
            {
                printf(
                    "%s %s\n"
                    "Copyright (C) 1992,2004 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "This is free software; see the source for copying conditions. There is NO\n" 
                    "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
                    Argh.ProgName().c_str(), VERSION
                      );
                return 0;
            }
            
            case 'h':
            {
                std::printf(   
                    "65816 linker\n"
                    "Copyright (C) 1992,2004 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\nUsage: %s [<option> [<...>]] <file> [<...>]\n"
                    "\nLinks O65 and IPS files together and produces an IPS file.\n"
                    "\nOptions:\n",
                    Argh.ProgName().c_str());
                Argh.ListOptions();
                std::printf("\nNo warranty whatsoever.\n");
                return 0;
            }

            case 'o':
            {
                outfn = Argh.GetString();
                if(outfn != "-")
                {
                    if(output)
                    {
                        std::fclose(output);
                    }
                    output = std::fopen(outfn.c_str(), "wb");
                    if(!output)
                    {
                        std::perror(outfn.c_str());
                ErrorExit:
                        if(output)
                        {
                            fclose(output);
                            remove(outfn.c_str());
                        }
                        return -1;
                    }
                }
                break;
            }

            default: 
                files.push_back(Argh.GetString());
        }
    }
EndArgList:  
    if(!Argh.ok()) goto ErrorExit;
    if(!files.size())
    {
        fprintf(stderr, "Error: Link what? See %s --help\n",
            Argh.ProgName().c_str());
        goto ErrorExit;
    }
    
    
    O65linker linker;
    freespacemap freespace;
    
    // Assume everything is free space!
    for(unsigned page=0x00; page<=0x3F; ++page)
        freespace.Add(page, 0x0000, 0x10000);
    
    for(unsigned a=0; a<files.size(); ++a)
    {
        char Buf[5];
        FILE *fp = fopen(files[a].c_str(), "rb");
        if(!fp)
        {
            perror(files[a].c_str());
            continue;
        }
        
        fread(Buf, 1, 5, fp);
        if(!strncmp(Buf, "PATCH", 5))
        {
            linker.LoadIPSfile(fp, files[a]);
        }
        else
        {
            O65 tmp;
            tmp.Load(fp);
            linker.AddObject(tmp, files[a]);
        }
        fclose(fp);
    }
    
    /* Organize the code blobs */    
    vector<unsigned> o65sizes = linker.GetSizeList();
    vector<unsigned> o65addrs = linker.GetAddrList();
    unsigned ignore = 0;
    for(unsigned a=0; a<o65addrs.size(); ++a)
        if(o65addrs[a])
        {
            freespace.Del(o65addrs[a], o65sizes[a]);
            ++ignore;
        }

    vector<freespacerec> Organization(o65sizes.size() - ignore);
    for(unsigned a=0, b=0; a<o65sizes.size(); ++a)
    {
        if(!o65addrs[a]) Organization[b++].len = o65sizes[a];
    }

    freespace.OrganizeToAnyPage(Organization);

    for(unsigned a=0, b=0; a<o65addrs.size(); ++a) 
        if(!o65addrs[a])
        {
            o65addrs[a] = Organization[b++].pos | 0xC00000;
        }

    linker.PutAddrList(o65addrs);
    
    linker.Link();
    
    vector<unsigned> tmp;
    
    std::FILE* stream = output ? output : stdout;

    fprintf(stream, "PATCH");
    
    o65addrs = linker.GetAddrList();
    for(unsigned a=0; a<o65addrs.size(); ++a)
    {
        const vector<unsigned char>& code = linker.GetCode(a);
        
        unsigned char data[5];
        
        data[0] = (o65addrs[a] >>16) & 0x3F;
        data[1] = (o65addrs[a] >> 8) & 0xFF;
        data[2] = (o65addrs[a]     ) & 0xFF;
        data[3] = (code.size() >> 8) & 0xFF;
        data[4] = (code.size()     ) & 0xFF;
        
        fwrite(data, 1, 5, stream);
        fwrite(&code[0], 1, code.size(), stream);
    }

    fprintf(stream, "EOF");
    
    if(output) fclose(output);
    
    return 0;
}
